#include "Cv12_main.h"

//----------------------------------------
//----    STRUCTURES AND FUNCTIONS    ----
//----------------------------------------

BindlessPtrNV BindlessPtrNV::FromVertexBuffer(GLuint index, GLuint buffer, size_t offset)
{
	BindlessPtrNV ptr;

	// Get the address and the size of given buffer
	GLuint64 buffer_address;
	GLint buffer_size;
	glGetNamedBufferParameterui64vNV(buffer, GL_BUFFER_GPU_ADDRESS_NV, &buffer_address);
	glGetNamedBufferParameterivEXT(buffer, GL_BUFFER_SIZE, &buffer_size);
	
	// Set the parameters
	ptr.index = index;
	ptr.reserved = 0;
	ptr.address = buffer_address + GLuint64(offset);
	ptr.length = GLuint64(buffer_size) - GLuint64(offset);

	return ptr;
}

BindlessPtrNV BindlessPtrNV::FromIndexBuffer(GLuint buffer)
{
	return FromVertexBuffer(0, buffer, 0);
}

DrawElementsIndirectCommand DrawElementsIndirectCommand::New(GLuint count)
{
	DrawElementsIndirectCommand cmd;
	cmd.count = count;
	cmd.instanceCount = 1;
	cmd.firstIndex = 0;
	cmd.baseVertex = 0;
	cmd.baseInstance = 0;
	return cmd;
}

//---------------------
//----    SCENE    ----
//---------------------

/// Reloads all shaders
void reload_shaders()
{
	object_program.Init();
	object_program.AddVertexShader("Shaders/texture_vertex.glsl");
	object_program.AddFragmentShader("Shaders/texture_fragment.glsl");
	object_program.Link();

	cout << "Shaders are reloaded" << endl;
}

/// Initializes all objects of our scene
void init_scene()
{
	//----------------------------------------------
	//--  Display some information about the OpenGL

	PrintOpenGLInfo();

	cout << "Extension GL_ARB_buffer_storage is " << (IsOpenGLExtensionPresent("GL_ARB_buffer_storage") ? "" : "not ") << "present" << endl;
	cout << "Extension GL_EXT_direct_state_access is " << (IsOpenGLExtensionPresent("GL_EXT_direct_state_access") ? "" : "not ") << "present" << endl;
	cout << "Extension GL_NV_shader_buffer_load is " << (IsOpenGLExtensionPresent("GL_NV_shader_buffer_load") ? "" : "not ") << "present" << endl;
	cout << "Extension GL_NV_vertex_buffer_unified_memory is " << (IsOpenGLExtensionPresent("GL_NV_vertex_buffer_unified_memory") ? "" : "not ") << "present" << endl;
	cout << "Extension GL_NV_bindless_multi_draw_indirect is " << (IsOpenGLExtensionPresent("GL_NV_bindless_multi_draw_indirect") ? "" : "not ") << "present" << endl;
	cout << "Extension GL_ARB_shader_draw_parameters is " << (IsOpenGLExtensionPresent("GL_ARB_shader_draw_parameters") ? "" : "not ") << "present" << endl;
	cout << endl;

	// Check whether given setup is correct
	if (TASK_ONE_METHOD != TASK_ONE_METHOD_OLD_WAY)
	{
		// Other Task 1 methods need GL_ARB_buffer_storage extension to work
		if (!IsOpenGLExtensionPresent("GL_ARB_buffer_storage"))
		{
			cout << "Application needs GL_ARB_buffer_storage extension to run. Application will exit." << endl;
			exit(0);
		}
	}
	if (TASK_TWO_METHOD == TASK_TWO_METHOD_USE)
	{
		// Task 2 needs these extensions to work:
		//		GL_EXT_direct_state_access
		//		GL_NV_shader_buffer_load
		//		GL_NV_vertex_buffer_unified_memory
		//		GL_NV_bindless_multi_draw_indirect
		//		GL_ARB_shader_draw_parameters
		if (!IsOpenGLExtensionPresent("GL_EXT_direct_state_access") ||
			!IsOpenGLExtensionPresent("GL_NV_shader_buffer_load") ||
			!IsOpenGLExtensionPresent("GL_NV_vertex_buffer_unified_memory") ||
			!IsOpenGLExtensionPresent("GL_NV_bindless_multi_draw_indirect") ||
			!IsOpenGLExtensionPresent("GL_ARB_shader_draw_parameters"))
		{
			cout << "Application needs GL_EXT_direct_state_access, GL_NV_shader_buffer_load, GL_NV_vertex_buffer_unified_memory, "
				"GL_NV_bindless_multi_draw_indirect, and GL_ARB_shader_draw_parameters extensions to run. Application will exit." << endl;
			exit(0);
		}
	}

	//----------------------------------------------
	//--  Load shaders

	reload_shaders();
	
	//----------------------------------------------
	//--  Prepare materials

	// Prepare materials for the random scene and for the floor
	RedMaterial		= PhongMaterial::CreateBasicMaterial(glm::vec3(1.0f, 0.0f, 0.0f), true, 200.0f, 1.0f);
	GreenMaterial	= PhongMaterial::CreateBasicMaterial(glm::vec3(0.0f, 1.0f, 0.0f), true, 200.0f, 1.0f);
	BlueMaterial	= PhongMaterial::CreateBasicMaterial(glm::vec3(0.0f, 0.0f, 1.0f), true, 200.0f, 1.0f);
	CyanMaterial	= PhongMaterial::CreateBasicMaterial(glm::vec3(0.0f, 1.0f, 1.0f), true, 200.0f, 1.0f);
	MagentaMaterial	= PhongMaterial::CreateBasicMaterial(glm::vec3(1.0f, 0.0f, 1.0f), true, 200.0f, 1.0f);
	YellowMaterial	= PhongMaterial::CreateBasicMaterial(glm::vec3(1.0f, 1.0f, 0.0f), true, 200.0f, 1.0f);
	WhiteMaterial	= PhongMaterial::CreateBasicMaterial(glm::vec3(1.0f, 1.0f, 1.0f), true, 200.0f, 1.0f);
	FloorMaterial	= PhongMaterial::CreateBasicMaterial(glm::vec3(0.7f, 0.7f, 0.7f), true, 200.0f, 1.0f);

	// Add the materials into our list of materials (we do not add the material for the floor, only the basic colors)
	Colors.push_back(&RedMaterial);
	Colors.push_back(&GreenMaterial);
	Colors.push_back(&BlueMaterial);
	Colors.push_back(&CyanMaterial);
	Colors.push_back(&MagentaMaterial);
	Colors.push_back(&YellowMaterial);
	Colors.push_back(&WhiteMaterial);

	//----------------------------------------------
	//--  Prepare geometries
	
	geom_cube = CreateCube();
	geom_sphere = CreateSphere();
	geom_torus = CreateTorus();
	geom_cylinder = CreateCylinder();
	geom_capsule = CreateCapsule();
	geom_teapot = CreateTeapot();

	if (TASK_TWO_METHOD == TASK_TWO_METHOD_USE)
	{
		// Make all vertex buffers and index buffers of all geometries "resident",
		// which allows us to bind them to OpenGL using GPU addresses to their data.
		glMakeNamedBufferResidentNV(geom_cube.VertexBuffers[0], GL_READ_ONLY);
		glMakeNamedBufferResidentNV(geom_sphere.VertexBuffers[0], GL_READ_ONLY);
		glMakeNamedBufferResidentNV(geom_torus.VertexBuffers[0], GL_READ_ONLY);
		glMakeNamedBufferResidentNV(geom_cylinder.VertexBuffers[0], GL_READ_ONLY);
		glMakeNamedBufferResidentNV(geom_capsule.VertexBuffers[0], GL_READ_ONLY);
		glMakeNamedBufferResidentNV(geom_teapot.VertexBuffers[0], GL_READ_ONLY);

		glMakeNamedBufferResidentNV(geom_cube.IndexBuffer, GL_READ_ONLY);
		glMakeNamedBufferResidentNV(geom_sphere.IndexBuffer, GL_READ_ONLY);
		glMakeNamedBufferResidentNV(geom_torus.IndexBuffer, GL_READ_ONLY);
		glMakeNamedBufferResidentNV(geom_cylinder.IndexBuffer, GL_READ_ONLY);
		glMakeNamedBufferResidentNV(geom_capsule.IndexBuffer, GL_READ_ONLY);
		glMakeNamedBufferResidentNV(geom_teapot.IndexBuffer, GL_READ_ONLY);
	}

	// Add the geometries into our list of geometries. Also add matrices that describe
	// how to transform the objects so that they "sit" on xz plane.
	Geometries.push_back(make_pair(&geom_cube,
		glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 1.0f, 0.0f))));
	Geometries.push_back(make_pair(&geom_sphere,
		glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 1.0f, 0.0f))));
	Geometries.push_back(make_pair(&geom_torus,
		glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 1.5f, 0.0f)) *
		glm::rotate(glm::mat4(1.0f), glm::radians(90.0f), glm::vec3(1.0f, 0.0f, 0.0f))));
	Geometries.push_back(make_pair(&geom_cylinder,
		glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 0.5f, 0.0f)) *
		glm::rotate(glm::mat4(1.0f), glm::radians(90.0f), glm::vec3(1.0f, 0.0f, 0.0f))));
	Geometries.push_back(make_pair(&geom_capsule,
		glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 0.5f, 0.0f)) *
		glm::rotate(glm::mat4(1.0f), glm::radians(90.0f), glm::vec3(1.0f, 0.0f, 0.0f))));
	Geometries.push_back(make_pair(&geom_teapot,
		glm::mat4(1.0f)));

	if (TASK_TWO_METHOD == TASK_TWO_METHOD_USE)
	{
		// Task 2: Create a VAO object that contains only the format of vertex atributes, not the buffers.
		//		Hints:
		//			glVertexAttribFormatNV sets parameters of vertex attributes, just like glVertexAttribPointer,
		//				but unlike glVertexAttribPointer, it does not store the pointer to the data.
		//			Parameters of glVertexAttribPointer are (index, size, type, normalized, stride, pointer)
		//			Parameters of glVertexAttribFormatNV are (index, size, type, normalized, stride)
		//			See function CreateCube/Sphere/Torus/Cylinder/Capsule/Teapot, find glVertexAttribPointer that
		//				set parameters of positions and normals, and set these parameters to glVertexAttribFormatNV
		//			Notice that these parameters are the same for all geometries. If they weren't, we couldn't do this.
		glGenVertexArrays(1, &VertexFormat_VAO);
		glBindVertexArray(VertexFormat_VAO);

		glEnableVertexAttribArray(DEFAULT_POSITION_LOC);
		//glVertexAttribFormatNV(DEFAULT_POSITION_LOC, ...);
			
		glEnableVertexAttribArray(DEFAULT_NORMAL_LOC);
		//glVertexAttribFormatNV(DEFAULT_NORMAL_LOC, ...);
		
		glBindVertexArray(0);
	}

	//----------------------------------------------
	//--  Positions of objects in the scene

	// We create a grid of random objects with random materials.
	srand(12345);
	glm::vec3 xyz_start = glm::vec3(-30.0f, 0.0f, -30.0f);
	glm::vec3 xyz_spacing = glm::vec3(2.0f, 0.0f, 2.0f);
	glm::ivec3 xyz_grid_size = glm::ivec3(30, 1, 30);
	size_t number_of_objects_in_scene = xyz_grid_size.x * xyz_grid_size.y * xyz_grid_size.z + 1;

	// Allocate the buffer with positions
	PositionsOfObjectsInScene.Init(number_of_objects_in_scene, GL_SHADER_STORAGE_BUFFER);
	// Allocate the buffer with materials
	MaterialsOfObjectsInScene.Init(number_of_objects_in_scene, GL_SHADER_STORAGE_BUFFER);
	// Allocate the array with draw commands (for Task 2)
	vector<DrawElementsIndirectBindlessCommandNV> draw_commands(number_of_objects_in_scene);

	for (int z = 0; z < xyz_grid_size.z; z++)
	for (int y = 0; y < xyz_grid_size.y; y++)
	for (int x = 0; x < xyz_grid_size.x; x++)
	{
		glm::ivec3 xyz = glm::ivec3(x, y, z);
		int model_idx = z*xyz_grid_size.x*xyz_grid_size.y + y*xyz_grid_size.x + x;

		// Choose the geometry randomly
		int geometry = rand() % int(Geometries.size());

		// Choose the material randomly
		int material = rand() % int(Colors.size());
		// Set the material
		MaterialsOfObjectsInScene.SetMaterial(model_idx, *Colors[material]);

		// Rotate the object randomly around the y axis
		glm::mat4 rotation = glm::rotate(glm::mat4(1.0f), float(rand()) / float(RAND_MAX) * float(M_PI) * 2.0f, glm::vec3(0.0f, 1.0f, 0.0f));
		// Place the object at the proper place in the grid
		glm::mat4 translation = glm::translate(glm::mat4(1.0f), xyz_start + xyz_spacing * glm::vec3(xyz));
		// Compute the model matrix
		glm::mat4 obj_matrix = translation * rotation * Geometries[geometry].second;
		// Set the model matrix
		PositionsOfObjectsInScene.SetMatrix(model_idx, obj_matrix);

		// Create the scene object and add it into the list
		SceneObject scene_object;
		scene_object.geometry = Geometries[geometry].first;
		scene_object.idx = model_idx;
		ObjectsInScene.push_back(scene_object);

		if (TASK_TWO_METHOD == TASK_TWO_METHOD_USE)
		{
			// Task 2: Create the draw command for scene_object.geometry, i.e. set all parameters of draw_commands[model_idx]
			//		using values in scene_object.geometry
			//			- Use DrawElementsIndirectCommand::New to set 'cmd'
			//				- See glDrawElements in method Geometry::Draw (PV227_Basics.cpp) to find the value for parameter 'count'
			//			- Set 'reserved' to zero (it is already done)
			//			- Use BindlessPtrNV::FromIndexBuffer to set indexBuffer. The buffer to set is the Geometry::IndexBuffer.
			//			- Use BindlessPtrNV::FromVertexBuffer to set vertexBuffers[0] and vertexBuffers[1]:
			//				- Their index is 'DEFAULT_POSITION_LOC' and 'DEFAULT_NORMAL_LOC'
			//				- Their buffer is Geometry::VertexBuffers[0] for both attributes (data of both attributes are in the same buffer)
			//				- Their offsets are in CreateCube/Sphere/..., it is the last parameter of glVertexAttribPointer of corresponding
			//					attribute, but as an integer, not as pointer.
			//draw_commands[model_idx].cmd = DrawElementsIndirectCommand::New(...);
			draw_commands[model_idx].reserved = 0;
			//draw_commands[model_idx].indexBuffer = BindlessPtrNV::FromIndexBuffer(...);
			//draw_commands[model_idx].vertexBuffers[0] = BindlessPtrNV::FromVertexBuffer(...);
			//draw_commands[model_idx].vertexBuffers[1] = BindlessPtrNV::FromVertexBuffer(...);
		}
	}

	// Add the floor object
	PositionsOfObjectsInScene.SetMatrix(number_of_objects_in_scene - 1, glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, -0.1f, 0.0f)) *
		glm::scale(glm::mat4(1.0f), glm::vec3(40.0f, 0.1f, 40.0f)));
	MaterialsOfObjectsInScene.SetMaterial(number_of_objects_in_scene - 1, FloorMaterial);

	SceneObject floor_object;
	floor_object.geometry = &geom_cube;
	floor_object.idx = int(number_of_objects_in_scene - 1);
	ObjectsInScene.push_back(floor_object);

	if (TASK_TWO_METHOD == TASK_TWO_METHOD_USE)
	{
		// Task 2: Create the draw command for floor_object.geometry, i.e. set all parameters of draw_commands[floor_object.idx]
		//		Hint: Use the same way as for other objects in the scene
		//draw_commands[floor_object.idx].cmd = DrawElementsIndirectCommand::New(...);
		draw_commands[floor_object.idx].reserved = 0;
		//draw_commands[floor_object.idx].indexBuffer = BindlessPtrNV::FromIndexBuffer(...);
		//draw_commands[floor_object.idx].vertexBuffers[0] = BindlessPtrNV::FromVertexBuffer(...);
		//draw_commands[floor_object.idx].vertexBuffers[1] = BindlessPtrNV::FromVertexBuffer(...);
	}

	if (TASK_TWO_METHOD == TASK_TWO_METHOD_USE)
	{
		// Creates a OpenGL buffer with all draw commands (for Task 2)
		glGenBuffers(1, &DrawCommandsBuffer);
		glBindBuffer(GL_DRAW_INDIRECT_BUFFER, DrawCommandsBuffer);
		glBufferData(GL_DRAW_INDIRECT_BUFFER, sizeof(DrawElementsIndirectBindlessCommandNV) * draw_commands.size(),
			draw_commands.data(), GL_STATIC_DRAW);
		glBindBuffer(GL_DRAW_INDIRECT_BUFFER, 0);
	}

	// Uploads the positions to GPU
	PositionsOfObjectsInScene.UpdateOpenGLData();
	// Uploads the materials to GPU
	MaterialsOfObjectsInScene.UpdateOpenGLData();

	//----------------------------------------------
	//--  Prepare all cameras
	
	// Create the basic camera through which we look into the scene
	the_camera.SetCenter(glm::vec3(0.0f));
	the_camera.SetCamera(-0.5f, -0.5f, 15.0f);

	if (TASK_ONE_METHOD == TASK_ONE_METHOD_OLD_WAY)
	{
		// The old way, as we did in all previous lessons
		glGenBuffers(1, &CameraBuffer);
		glBindBuffer(GL_UNIFORM_BUFFER, CameraBuffer);
		glBufferData(GL_UNIFORM_BUFFER, sizeof(CameraData_UBO::SingleCameraData), nullptr, GL_STATIC_DRAW);
		glBindBuffer(GL_UNIFORM_BUFFER, 0);
	}
	else if (TASK_ONE_METHOD == TASK_ONE_METHOD_NEW_WAY_NEW_STORAGE)
	{
		// The new way, but with a single camera
		glGenBuffers(1, &CameraBuffer);
		glBindBuffer(GL_UNIFORM_BUFFER, CameraBuffer);
		glBufferStorage(GL_UNIFORM_BUFFER, sizeof(CameraData_UBO::SingleCameraData), nullptr, GL_DYNAMIC_STORAGE_BIT);
		glBindBuffer(GL_UNIFORM_BUFFER, 0);
	}
	else if (TASK_ONE_METHOD == TASK_ONE_METHOD_NEW_WAY_NEW_UPDATE_INCORRECT)
	{
		// The new way, set to be updated by directly from CPU without CPU-GPU copies
		glGenBuffers(1, &CameraBuffer);
		glBindBuffer(GL_UNIFORM_BUFFER, CameraBuffer);
		// Set additional flags so that we can access the data in the buffer from the CPU
		glBufferStorage(GL_UNIFORM_BUFFER, sizeof(CameraData_UBO::SingleCameraData), nullptr, GL_DYNAMIC_STORAGE_BIT |
			GL_MAP_WRITE_BIT | GL_MAP_PERSISTENT_BIT | GL_MAP_COHERENT_BIT);
		// Map the buffer to get CPU pointer of the data of the buffer. See that thanks to GL_MAP_PERSISTENT_BIT,
		// the buffer can be permanently mapped - it does not need to be unmapped.
		CameraBufferData = static_cast<CameraData_UBO::SingleCameraData *>(glMapBufferRange(GL_UNIFORM_BUFFER, 0, sizeof(CameraData_UBO::SingleCameraData),
			GL_MAP_WRITE_BIT | GL_MAP_PERSISTENT_BIT | GL_MAP_COHERENT_BIT));
		glBindBuffer(GL_UNIFORM_BUFFER, 0);
	}
	else if (TASK_ONE_METHOD == TASK_ONE_METHOD_NEW_WAY_NEW_UPDATE_CORRECT)
	{
		// Task 1: Create a copy of TASK_ONE_METHOD_NEW_WAY_NEW_UPDATE (the one above) and change it to
		//		work correctly and without waiting at the synchronization.
		//		Hint: Create multiple buffers, create a circular buffer
	}

	//----------------------------------------------
	//--  Prepare the lights

	PhongLights_ubo.Init();
	PhongLights_ubo.SetGlobalAmbient(glm::vec3(0.2f));
	PhongLights_ubo.PhongLights.clear();
	PhongLights_ubo.PhongLights.push_back(
		PhongLight::CreatePointLight(glm::vec3(+20.0f, 5.0f, +20.0f), glm::vec3(0.0f), glm::vec3(1.0f), glm::vec3(1.0f), 0.5f, 0.1f, 0.0f));
	PhongLights_ubo.PhongLights.push_back(
		PhongLight::CreatePointLight(glm::vec3(+20.0f, 5.0f, -20.0f), glm::vec3(0.0f), glm::vec3(1.0f), glm::vec3(1.0f), 0.5f, 0.1f, 0.0f));
	PhongLights_ubo.PhongLights.push_back(
		PhongLight::CreatePointLight(glm::vec3(-20.0f, 5.0f, +20.0f), glm::vec3(0.0f), glm::vec3(1.0f), glm::vec3(1.0f), 0.5f, 0.1f, 0.0f));
	PhongLights_ubo.PhongLights.push_back(
		PhongLight::CreatePointLight(glm::vec3(-20.0f, 5.0f, -20.0f), glm::vec3(0.0f), glm::vec3(1.0f), glm::vec3(1.0f), 0.5f, 0.1f, 0.0f));
	PhongLights_ubo.UpdateOpenGLData();

	//----------------------------------------------
	//--  Miscelaneous

	// Create the query object
	glGenQueries(1, &RenderTimeQuery);
	query_waiting = false;

	// Turn off V-Sync
	SetVSync(false);

	// Initialize fence objects
	if ((TASK_ONE_METHOD == TASK_ONE_METHOD_OLD_WAY) ||
		(TASK_ONE_METHOD == TASK_ONE_METHOD_NEW_WAY_NEW_STORAGE))
	{
		// No fences needed
	}
	else if (TASK_ONE_METHOD == TASK_ONE_METHOD_NEW_WAY_NEW_UPDATE_INCORRECT)
	{
		// The fence is created when used. We set it to nullptr here to indicate
		// it hasn't been created yet.
		EndFence = nullptr;
	}
	else if (TASK_ONE_METHOD == TASK_ONE_METHOD_NEW_WAY_NEW_UPDATE_CORRECT)
	{
		// Task 1: Initialize all fence objects that you will need
	}
}

/// Updates the scene: performs animations, updates the data of the buffers, etc.
void update_scene(int app_time_diff_ms)
{
	//----------------------------------------------
	//--  Data of the main camera
	
	CameraData_UBO::SingleCameraData cam_data;
	cam_data.projection = glm::perspective(glm::radians(45.0f), float(win_width) / float(win_height), 0.1f, 1000.0f);
	cam_data.projection_inv = glm::inverse(cam_data.projection);
	cam_data.view = the_camera.GetViewMatrix();
	cam_data.view_inv = the_camera.GetViewInvMatrix();
	cam_data.view_it = glm::mat3x4(glm::transpose(glm::inverse(glm::mat3(cam_data.view))));
	cam_data.eye_position = glm::vec4(the_camera.GetEyePosition(), 1.0f);

	if ((TASK_ONE_METHOD == TASK_ONE_METHOD_OLD_WAY) ||
		(TASK_ONE_METHOD == TASK_ONE_METHOD_NEW_WAY_NEW_STORAGE))
	{
		// This is how to update the data in old way and in new way when the buffer is not mapped - call glBufferSubData
		glBindBuffer(GL_UNIFORM_BUFFER, CameraBuffer);
		glBufferSubData(GL_UNIFORM_BUFFER, 0, sizeof(CameraData_UBO::SingleCameraData), &cam_data);
		glBindBuffer(GL_UNIFORM_BUFFER, 0);
	}
	else if (TASK_ONE_METHOD == TASK_ONE_METHOD_NEW_WAY_NEW_UPDATE_INCORRECT)
	{
		// Task 1: This is not a correct solution, it produces artefacts. It only shows how to update
		//			the data and how to work with sync objects.
		//			There is nothing to add, add the code into the following part (TASK_ONE_METHOD_NEW_WAY_NEW_UPDATE_CORRECT)

		// This is how to update the data - simply store the data into the buffer using the CPU pointer
		//		This way, you can rewrite the data that are still in use (don't forget rendering is asynchronous),
		//		you as the programmer are the one that must ensure noone uses the old data (e.g. by checking fences).
		*CameraBufferData = cam_data;

		// This is how to work with OpenGL sync objects
		if (EndFence)			// If the fence object was created
		{
			// Get the status of the fence, see the documentation of glClientWaitSync for all possible statuses
			GLenum status = glClientWaitSync(EndFence, 0, 0);
			if (status == GL_ALREADY_SIGNALED)
			{
				// The fence is signaled, which means all OpenGL operations up to the fence completed.
				// This means the update a few lines above did not overwriten any data that would be still in use.
			}
			else
			{
				// The fence is not signaled, which means OpenGL operations have not completed yet.
				// This means the update a few lines above could overwrite some data that are still in use.
				// We do not properly solve this situation in this lecture, we just notify the user.
				cout << "Data were written too soon" << endl;
			}

			// Delete the fence object. Fence objects cannot be reused, they are deleted when no more needed
			glDeleteSync(EndFence);
			EndFence = nullptr;
		}
	}
	else if (TASK_ONE_METHOD == TASK_ONE_METHOD_NEW_WAY_NEW_UPDATE_CORRECT)
	{
		// Task 1: Update the data in the buffer correctly and without waiting at the synchronization.
		//		Hint: See the part 'TASK_ONE_METHOD_NEW_WAY_NEW_UPDATE_INCORRECT' above to see how to update the
		//			data of the camera and how to work with GLsync objects.
	}
}

/// Renders the whole frame
void render_scene()
{
	// Start measuring the elapsed time
	if (!query_waiting)
		glBeginQuery(GL_TIME_ELAPSED, RenderTimeQuery);

	// Render into the main window
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
	glViewport(0, 0, win_width, win_height);

	// Clear the framebuffer, clear the color and depth
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glClearDepth(1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);

	// Set the data of the camera
	if ((TASK_ONE_METHOD == TASK_ONE_METHOD_OLD_WAY) ||
		(TASK_ONE_METHOD == TASK_ONE_METHOD_NEW_WAY_NEW_STORAGE) ||
		(TASK_ONE_METHOD == TASK_ONE_METHOD_NEW_WAY_NEW_UPDATE_INCORRECT))
	{
		// No matter which type of buffer we used, we bind them in the same way as we always did
		glBindBufferBase(GL_UNIFORM_BUFFER, DEFAULT_CAMERA_BINDING, CameraBuffer);
	}
	else if (TASK_ONE_METHOD == TASK_ONE_METHOD_NEW_WAY_NEW_UPDATE_CORRECT)
	{
		// Task 1: Bind proper buffer with the data for the current frame
	}
	// Set the data of the lights
	PhongLights_ubo.BindBuffer(DEFAULT_LIGHTS_BINDING);
	// Set the buffer with positions of all objects in the scene
	PositionsOfObjectsInScene.BindBuffer(DEFAULT_OBJECT_BINDING);
	// Set the buffer with materials of all objects in the scene
	MaterialsOfObjectsInScene.BindBuffer(DEFAULT_MATERIAL_BINDING);

	// Render all objects in the scene
	if (object_program.IsValid())
	{
		if (TASK_TWO_METHOD == TASK_TWO_METHOD_USE)
		{
			// Task 2: No code to add. Just notice that we set *one* VAO, *one* buffer with commands,
			//		 and call *one* Draw function to draw all objects in the whole scene

			object_program.Use();
			glBindVertexArray(VertexFormat_VAO);

			// See the specification of GL_NV_vertex_buffer_unified_memory extension for more info on these two states
			glEnableClientState(GL_VERTEX_ATTRIB_ARRAY_UNIFIED_NV);
			glEnableClientState(GL_ELEMENT_ARRAY_UNIFIED_NV);

			// See https://www.opengl.org/wiki/Vertex_Rendering#Indirect_rendering for more info on indirect rendering
			// See the specification of GL_NV_bindless_multi_draw_indirect extension for more info on glMultiDrawElementsIndirectBindlessNV
			glBindBuffer(GL_DRAW_INDIRECT_BUFFER, DrawCommandsBuffer);
			glMultiDrawElementsIndirectBindlessNV(GL_TRIANGLE_STRIP, GL_UNSIGNED_INT, nullptr, ObjectsInScene.size(),
				sizeof(DrawElementsIndirectBindlessCommandNV), 2);
			glBindBuffer(GL_DRAW_INDIRECT_BUFFER, 0);

			glDisableClientState(GL_VERTEX_ATTRIB_ARRAY_UNIFIED_NV);
			glDisableClientState(GL_ELEMENT_ARRAY_UNIFIED_NV);
		}
		else
		{
			// The old method - for each object separately, we set its VAO and we call its draw function
			object_program.Use();
		
			for (auto iter = ObjectsInScene.begin(); iter != ObjectsInScene.end(); ++iter)
			{
				iter->geometry->BindVAO();

				object_program.Uniform1i("model_idx", iter->idx);		// 'model_idx' is used to get object's position and material
				iter->geometry->Draw();
			}
		}
	}

	// Reset the VAO and the program
	glBindVertexArray(0);
	glUseProgram(0);
	glBindFramebuffer(GL_FRAMEBUFFER, 0);

	// Stop measuring the elapsed time
	if (!query_waiting)
		glEndQuery(GL_TIME_ELAPSED);

	// Evaluate the render time query
	GLint ready;
	glGetQueryObjectiv(RenderTimeQuery, GL_QUERY_RESULT_AVAILABLE, &ready);
	if (ready != GL_FALSE)
	{
		// The query is ready. Evaluate it and prepare to the next frame.
		GLuint64 render_time;
		glGetQueryObjectui64v(RenderTimeQuery, GL_QUERY_RESULT, &render_time);
		render_time_ms = float(render_time) * 1e-6f;
		query_waiting = false;
	}
	else
	{
		// The query is not ready. Do not use it next frame
		query_waiting = true;
	}

	// Set fence into OpenGL command queue to know when this frame is really processed by the GPU
	if ((TASK_ONE_METHOD == TASK_ONE_METHOD_OLD_WAY) ||
		(TASK_ONE_METHOD == TASK_ONE_METHOD_NEW_WAY_NEW_STORAGE))
	{
		// No fences needed
	}
	else if (TASK_ONE_METHOD == TASK_ONE_METHOD_NEW_WAY_NEW_UPDATE_INCORRECT)
	{
		// Create a fence object to store a time stamp of the 
		// Do not create fence object until is it processed
		if (!EndFence)
		{
			EndFence = glFenceSync(GL_SYNC_GPU_COMMANDS_COMPLETE, 0);
		}
	}
	else if (TASK_ONE_METHOD == TASK_ONE_METHOD_NEW_WAY_NEW_UPDATE_CORRECT)
	{
		// Task 1: Find a way how to use fences to properly update the buffers
	}

	// Evaluate FPS
	fps = GetFPS();

	glFlush();
	glutPostRedisplay();
}

//-------------------
//----    GUI    ----
//-------------------

/// Callback when user clicks on Reload button
void TW_CALL reload(void *)
{
	reload_shaders();
}

void init_gui()
{
	// Initialize AntTweakBar
	TwInit(TW_OPENGL_CORE, nullptr);
	TwWindowSize(win_width, win_height);
	TwGLUTModifiersFunc(glutGetModifiers);

	// Initialize GUI
	the_gui = TwNewBar("Parameters");
	TwDefine("Parameters size='270 150'");
	TwDefine("Parameters valueswidth=100");
	TwDefine("Parameters refresh=0.1");
	TwAddButton(the_gui, "Reload", reload, nullptr, nullptr);

	// Render time (ms)
	TwAddVarRO(the_gui, "Render time (ms)", TW_TYPE_FLOAT, &render_time_ms, nullptr);
	TwAddVarRO(the_gui, "FPS", TW_TYPE_FLOAT, &fps, nullptr);
}

//---------------------------
//----    APPLICATION    ----
//---------------------------

/// GLUT callback - when the user presses a key. See glutKeyboardFunc for more info
void on_keyboard_func(unsigned char key, int x, int y)
{
	// Inform AntTweakBar
	if (TwEventKeyboardGLUT(key, x, y))
		return;		// Already handled by AntTweakBar

	switch (key)
	{
	case 27:		// Escape key
		exit(0);
		break;
	}
}

/// GLUT callback - when the user presses a special key (like F1 key). See glutSpecialFunc for more info
void on_special_func(int key, int x, int y)
{
	// Inform AntTweakBar
	if (TwEventSpecialGLUT(key, x, y))
		return;		// Already handled by AntTweakBar
}

/// GLUT callback - when the user presses or releases a mouse button. See glutMouseFunc for more info
void on_mouse_func(int button, int state, int x, int y)
{
	// Inform AntTweakBar
	if (TwEventMouseButtonGLUT(button, state, x, y))
		return;		// Already handled by AntTweakBar

	the_camera.OnMouseFunc(button, state, x, y);
}

/// GLUT callback - when the user moves the mouse when holding a botton. See glutMotionFunc for more info
void on_motion_func(int x, int y)
{
	// Inform AntTweakBar
	if (TwEventMouseMotionGLUT(x, y))
		return;		// Already handled by AntTweakBar

	the_camera.OnMotionFunc(x, y);
}

/// GLUT callback - when the user moves the mouse when not holding a botton. See glutPassiveMotionFunc for more info
void on_passive_motion_func(int x, int y)
{
	// Inform AntTweakBar
	if (TwEventMouseMotionGLUT(x, y))
		return;		// Already handled by AntTweakBar
}

/// GLUT callback - when the window is resized. See glutReshapeFunc for more info
void on_reshape(int width, int height)
{
	win_width = width;
	win_height = height;
	
	glViewport(0, 0, win_width, win_height);
	
	// Inform AntTweakBar
	TwWindowSize(win_width, win_height);
}

/// GLUT callback - when our animation timer requests a new frame. See glutTimerFunc for more info
void on_timer(int)
{
	// Register the timer to be called again in 20 ms, and rerender the scene
	glutTimerFunc(20, on_timer, 0);
	glutPostRedisplay();
}

/// GLUT callback - when the window needs to be redrawn. See glutDisplayFunc for more info
void on_display()
{
	//--  Update all the data

	// Update the application time
	int current_glut_time = glutGet(GLUT_ELAPSED_TIME);
	int app_time_diff_ms = current_glut_time - last_glut_time;
	app_time_ms += app_time_diff_ms;
	last_glut_time = current_glut_time;

	// Update the scene
	update_scene(app_time_diff_ms);

	//--  Render the scene

	// Render the scene
	render_scene();
	
	// Render the GUI
	TwDraw();

	// Swaps the front and back buffer (double-buffering)
	glutSwapBuffers();
}

/// Callback function to be called when we make an error in OpenGL
void GLAPIENTRY simple_debug_callback(GLenum source, GLenum type, GLuint id, GLenum severity, GLsizei length, const char* message, const void* userParam)
{
	switch (type)			// Report only some errors
	{
	case GL_DEBUG_TYPE_ERROR:
	case GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR:
	case GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR:
		cout << message << endl;		// Put the breakpoint here
		return;
	default:
		return;
	}
}

/// C main function :-)
int main(int argc, char ** argv)
{
	// Initialize GLUT
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
	glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);		// Return from the main loop

	// Set OpenGL Context parameters
	glutInitContextVersion(4, 3);				// Use OpenGL 4.3
	glutInitContextProfile(GLUT_CORE_PROFILE);	// Use OpenGL core profile
	glutInitContextFlags(GLUT_DEBUG);			// Use OpenGL debug context

	// Initialize window
	glutInitWindowSize(win_width, win_height);
	glutCreateWindow("PV227 - Cv12: Modern OpenGL");

	// Initialize GLEW
	glewExperimental = GL_TRUE;
	glewInit();

	// Initialize DevIL library
	InitDevIL();

	// Set the OpenGL debug callback
	SetDebugCallback(simple_debug_callback);

	// Register GLUT callbacks
	glutDisplayFunc(on_display);
	glutReshapeFunc(on_reshape);
	glutTimerFunc(20, on_timer, 0);

	glutKeyboardFunc(on_keyboard_func);
	glutSpecialFunc(on_special_func);
	glutMouseFunc(on_mouse_func);
	glutMotionFunc(on_motion_func);
	glutPassiveMotionFunc(on_passive_motion_func);

	// Initialize OpenGL stuff
	init_gui();
	init_scene();

	// Everything is loaded and prepared, reset the application time
	last_glut_time = glutGet(GLUT_ELAPSED_TIME);
	app_time_ms = 0;

	// Run the main loop
	glutMainLoop();

	// Unload AntTweakBar
	TwTerminate();

	return 0;
}