// PV227, Lesson 12: Modern OpenGL
#pragma once

#include "../../Framework/PV227.h"

using namespace std;
using namespace PV227;

const int TASK_ONE_METHOD_OLD_WAY = 0;
const int TASK_ONE_METHOD_NEW_WAY_NEW_STORAGE = 1;
const int TASK_ONE_METHOD_NEW_WAY_NEW_UPDATE_INCORRECT = 2;
const int TASK_ONE_METHOD_NEW_WAY_NEW_UPDATE_CORRECT = 3;			// Task 1: Use this as TASK_ONE_METHOD
const int TASK_ONE_METHOD = TASK_ONE_METHOD_OLD_WAY;

const int TASK_TWO_METHOD_DO_NOT_USE = 0;
const int TASK_TWO_METHOD_USE = 1;									// Task 2: Use this as TASK_TWO_METHOD
const int TASK_TWO_METHOD = TASK_TWO_METHOD_DO_NOT_USE;

//----------------------------------------
//----    STRUCTURES AND FUNCTIONS    ----
//----------------------------------------

/// These three structures (BindlessPtrNV, DrawElementsIndirectCommand, DrawElementsIndirectBindlessCommandNV)
/// specifies the DrawElement rendering command and buffers to set, and allow us to rendering many objects
/// (even the whole scene) at once with one draw call.
///
/// BindlessPtrNV - specifies which buffer to set to which vertex shader input
/// DrawElementsIndirectCommand - specifies the parameters of one glDrawElements command (more precisely glDrawElementsInstancedBaseVertexBaseInstance)
/// DrawElementsIndirectBindlessCommandNV - specifies the rendering command, and vertex buffers and index buffer to use
struct BindlessPtrNV
{
	// See the specification of GL_NV_bindless_multi_draw_indirect extension for more info
	GLuint index;
	GLuint reserved;
	GLuint64 address;
	GLuint64 length;

	/// Creates a new BindlessPtrNV
	///		index	- Use index of vertex shader input for vertex buffer, use 0 for index buffer
	///		buffer	- Buffer in which the data is located
	///		offset	- offset in bytes where the data starts in the buffer
	static BindlessPtrNV FromVertexBuffer(GLuint index, GLuint buffer, size_t offset);
	/// Creates a new BindlessPtrNV, a version for index buffers
	static BindlessPtrNV FromIndexBuffer(GLuint buffer);
};
struct DrawElementsIndirectCommand
{
	// See glDrawElementsInstancedBaseVertexBaseInstance for more info
	GLuint count;
	GLuint instanceCount;
	GLuint firstIndex;
	GLint baseVertex;
	GLuint baseInstance;

	/// Creates a new draw command
	///		count	- Number of vertices to draw, in PV227::Geometry::DrawElementsCount
	static DrawElementsIndirectCommand New(GLuint count);
};
struct DrawElementsIndirectBindlessCommandNV
{
	DrawElementsIndirectCommand cmd;	// Draw command
	GLuint reserved;					// Set to 0
	BindlessPtrNV indexBuffer;			// Buffer with indexes
	BindlessPtrNV vertexBuffers[2];		// Buffers for each vertex shader input
};

//---------------------
//----    SCENE    ----
//---------------------

// Shader programs
ShaderProgram object_program;

// Geometries we use in this lecture
Geometry geom_cube;
Geometry geom_sphere;
Geometry geom_torus;
Geometry geom_cylinder;
Geometry geom_capsule;
Geometry geom_teapot;

// List of geometries which we choose for our scene
std::vector<std::pair<Geometry *, glm::mat4> > Geometries;

// VAO object that contains the format of all geometries, but not the buffers (for Task 2)
GLuint VertexFormat_VAO;

// A camera that allows us to look at the object from different views
SimpleCamera the_camera;

// Materials of objects
PhongMaterial RedMaterial;
PhongMaterial GreenMaterial;
PhongMaterial BlueMaterial;
PhongMaterial CyanMaterial;
PhongMaterial MagentaMaterial;
PhongMaterial YellowMaterial;
PhongMaterial WhiteMaterial;
PhongMaterial FloorMaterial;
// List of materials from which we choose for our scene
std::vector<PhongMaterial *> Colors;

/// Data of one object in the scene
struct SceneObject
{
	/// Geomety of the object
	Geometry *geometry;
	/// Index of the object which is used to get its material and position (they are stored all together for all objects in the scene)
	int idx;
};
/// List of objects in the scene
std::vector<SceneObject> ObjectsInScene;
/// Buffer with model matrices of all objects in the scene
ModelData_UBO PositionsOfObjectsInScene;
/// Buffer with materials of all objects in the scene
MaterialData_UBO MaterialsOfObjectsInScene;

// Buffer with draw commands (for Task 2)
GLuint DrawCommandsBuffer;

// Data of all our lights - positions, colors, etc.
PhongLightsData_UBO PhongLights_ubo;

GLuint CameraBuffer;									// Buffer with the data of a single camera
CameraData_UBO::SingleCameraData *CameraBufferData;		// Pointer to data of CameraBuffer buffer accessible by CPU 
GLsync EndFence;										// Fence at the end of the frame
// Task 1: Instead of using a single buffer for the data of the camera, use several buffers (and also
//			several pointers to their data, and several fences at the end of the frames).
//		- Hint: Duplicate the data above.
// Note: When we allocate for enough frames, we updates should stop being overwriten, because drivers are set not to be 
// too far from the CPU.

// OpenGL query object to get render time of one frame
GLuint RenderTimeQuery;
// When true, the data from the query are not read yet, so the query cannot be used to measure the frame
bool query_waiting;

// Functions that works with scene objects
void reload_shaders();
void init_scene();
void update_scene(int app_time_diff_ms);
void render_scene();

//-------------------
//----    GUI    ----
//-------------------

// Main AntTweakBar object
TwBar *the_gui;

// Variables that are changed with GUI
float render_time_ms;		// Read-only
float fps;					// Read-only

// Callbacks from the GUI
void TW_CALL reload(void *);

// Functions that works with GUI
void init_gui();

//---------------------------
//----    APPLICATION    ----

// Current window size
int win_width = 640;
int win_height = 480;

// Current time of the application in milliseconds, for animations
int app_time_ms = 0;
int last_glut_time = 0;
